<?php

/**
* Represents a Node Query
*
* Extends QueryTool
* Implements QueryToolInterface
*/
class NodeQueryTool extends QueryTool implements QueryToolInterface{
  
  public $type = 'node';
  
  /**
   * Load nodes into the QueryTool
   *
   * Implementation of QueryToolInterface::load
   */
  public function load($id, $method = 'one'){
    $invalid = array();
    
    if($node = node_load($id)){
      $this->items[$id] = $node;
    }
    else {
      $invalid = $id;
    }
    
    if($method != 'many'){
      if(isset($this->items)){
        $this->uid = 'node:'.md5(serialize(array_keys($this->items)));
      };
    }
    
    // Warn about any invalid NIDs provided
    if(isset($invalid[0])){
      watchdog('query_tools', 'Query tools tried to load the following invalid Node ID: '. $invalid, 'warning');
    }
  }
  
  /**
   * Iterator to load many ids in an array
   */
  // Iterate over NIDs and load these into the object
  function load_many($ids = array()){
    foreach($ids as $id){
      $this->load($id, 'many');
    }
    
    if(isset($this->items)){
      $this->uid = 'node:'.md5(serialize(array_keys($this->items)));
    };
  }
  
  /**
   * Retrieves a set of nodes
   *
   * Implements QueryToolInterface::find
   *
   * Apapted from http://drupal.org/node/360541#comment-3669676
   *
   * @param $param
   *   An array of conditions for core node fields to match against in the database query.
   *   Since we also have the CCK conditions, this parameter can also be NULL.
   * @param $cck_param (optional)
   *   An array of conditions for CCK fields to match against in the database query
   * @param $revision (optional)
   *   Which numbered revision to load. Defaults to the current version.
   *
   * @return
   *   A fully-populated array of node objects.
   */
  public function find() {
    $args = func_get_args();
    
    $parameters = $args[0]; 
    $cck_param = isset($args[1]) ? $args[1] : NULL;
    $revision = isset($args[2]) ? $args[2] : NULL;
    
    $this->query['# '] = array();

    /**
     * Process regular parameters
     */
    if (is_array($parameters)) {
      // Turn the conditions into a query.
      foreach ($parameters as $key => $value) {
        $cond[] = 'n.'. db_escape_table($key) ." = '%s'";
        $this->query['#arguments'][] = $value;
      }
      $cond = implode(' AND ', $cond);
    }

    /**
     * Process CCK parameters
     */
    $cck_tables = array();
    if (count($cck_param)) {
      // Turn the conditions into a query.
      $i = 1;
      foreach ($cck_param as $key => $value) {
        $field_name = 'field_'. str_replace('field_', '', $key); // ensure field has 'field_' at the beginning
        $db_info = content_database_info(content_fields($field_name));
        // If the field is from another table, increment sequence
        if(!isset($cck_tables[$db_info['table']])){
          $cck_tables[$db_info['table']] = 'cck'. $i;
        }
        $i++;
        // nodereference fields
        if($db_info['columns']['nid']){
          $cck_cond[] = $cck_tables[$db_info['table']] .'.'. db_escape_table($db_info['columns']['nid']['column']) ." = '%s'";
        } 
        // Normal fields
        elseif ($db_info['columns']['value']) {
          $cck_cond[] = $cck_tables[$db_info['table']] .'.'. db_escape_table($db_info['columns']['value']['column']) ." = '%s'";
        } 
        // User reference fields
        elseif ($db_info['columns']['uid']) {
          $cck_cond[] = $cck_tables[$db_info['table']] .'.'. db_escape_table($db_info['columns']['uid']['column']) ." = '%s'";
        }
        elseif ($db_info['columns'][$key]['column']) {
          $cck_cond[] = $cck_tables[$db_info['table']] .'.'. db_escape_table($db_info['columns'][$key]['column']) ." = '%s'";
        }
        $this->query['#arguments'][] = $value;
      }
      if (is_array($cck_cond)) {
        $cond .= (isset($cond) ? ' AND ' : ''). implode(' AND ', $cck_cond);
        foreach ($cck_tables as $table => $nick) {
          $cck_join .= ' INNER JOIN {'. $table .'} '. $nick .' ON '. $nick .'.nid = n.nid';
        }
      }
    }

    if ($revision) {
      array_unshift($this->query['#arguments'], $revision);
      $this->query['#sql'] = 'SELECT n.nid FROM {node} n INNER JOIN {node_revisions} r ON r.vid = n.vid'. $cck_join .' AND r.vid = %d WHERE '. $cond;
      $nodes = db_query($this->query['#sql'], $this->query['#arguments']);
    }
    else {
      $this->query['#sql'] = 'SELECT n.nid FROM {node} n INNER JOIN {node_revisions} r ON r.vid = n.vid'. $cck_join .' WHERE '. $cond;
      $nodes = db_query($this->query['#sql'], $this->query['#arguments']);
    }

    /**
     * Iterate over retrieved node values, and give node_load an opportunity 
     * to add more fields.
     */
    while ($node = db_fetch_object($nodes)) {
      if ($node && $node->nid) {
        $node = object_cache_get('node', $node->nid);
        if ($extra = node_invoke($node, 'load')) {
        foreach ($extra as $key => $value) {
            $node->$key = $value;
          }
        }
        if ($extra = node_invoke_nodeapi($node, 'load')) {
          foreach ($extra as $key => $value) {
            $node->$key = $value;
          }
        }
        $results[$node->nid] = $node;
      }
    }
    // Generate a unique key for this set of IDs, if IDs are provided
    if(isset($results)){
      $this->uid = query_tools_uid($results);
      $this->items = $results;
      $this->cache_set();
    }

    return $results;
  }
  
  
  /**
   * Implements QueryToolsInterface::update
   *
   * @param $fields
   *    An array of fields with values to update. e.g.
   *    array(
   *      'field_my_field' => array(
   *        '#type' => 'cck',
   *        '#values' => array(
   *          '0' => array(
   *            'value' => 'Value to insert',
   *          )
   *        )
   *      )
   *    )
   */
  function update($fields){

    // @todo autodetect the field type
    foreach($fields as $field => $data){
      switch ($data['#type']){
        case 'cck':
          $this->set_field($field, $data['#values']);
          break;
        default:
          $this->set_property($field, $data['#values']);
          break;
      } 
    }

    $result = $this->save();

    return $result;
  }
  
  /**
   * Update a field value
   *
   * This will set the field value on EVERY node referenced in this object.
   *
   * @param $field
   *    Name of the field to update
   * @param $value
   *    An array of values to set the field too
   */
  function set_field($field, $values){
    $new_values = array();
    $field_info = content_fields($field);
    $db_info = content_database_info($field);
    
    if(isset($this->items)){
      foreach($this->items as $key => $node){
        
        $this->items[$key]->$field = $values;
        $this->state_dirty($key);
      }
    }
    
    
  }
  
  /**
   * Save the node
   * 
   * Implements QueryToolInterface::save
   *   
   * @return
   *    A list of saved items if all of the nodes are saved correctly or FALSE
   */
  function save(){
    $cache = FALSE;
    $ret_items = array();
    
    if(isset($this->items)){
      foreach($this->items as $key => $node){
        if($node->query_tools['#state'] == 'dirty'){
          if ($node = node_submit($node)) {
          	node_save($node);
          	watchdog('query_tools', t("Node ".$node->title." updated correctly"));
          	$this->items[$key]->nid = $node->nid;
          	$this->state_clean($key);
          	$ret_items[$key] = $node;
          } else {
          	watchdog('query_tools', t("Node ".$node->title." updated incorrectly"), "error");
          	$this->state_dirty($key);
          }
          if($node->query_tools['#new'] == TRUE){
            $cache = TRUE;
          }
        }
      }
    }
    
    if($this->state() == 'clean'){
      if(isset($cache)){
        $this->cache_set();
      }
      return $ret_items;
    }
    else{
      drupal_set_message('Query tools could not save all nodes. Some data may be lost.');
    }
    
    return FALSE;
  }
  
  /**
   * Intialise a node object with sensible defaults
   *
   * @note
   *    This is not currently used.
   */
  function node_init(){
    $node = new stdClass();

    //Main Node Fields
    $node->name = 'Node name';
    $node->title = 'Node title';
    $node->body = '';
    $node->type = 'page'; //This can be any node type
    $node->created = time();
    $node->changed = $node->created;
    $node->promote = 0; // Display on front page ? 1 : 0
    $node->sticky = 0;  // Display top of page ? 1 : 0
    $node->format = 2;  // 1:Filtered HTML, 2: Full HTML
    $node->status = 1;   // Published ? 1 : 0
    $node->language = 'en';

    return $node;
  }
  
  /**
   * Create a new node record
   */
  public function create_node($args){
    $node = new stdClass();
    
    if (empty($args['user'])){
      global $user;

      //Main Node Fields
      $node->name = $user->name;
      $node->uid = $user->uid;
    }
    else {
      $node->name = $args['user']->name;
      $node->uid = $args['user']->uid;
    }
    
    $node->title = $args['title'];
    $node->type = $args['type']; //This can be any node type
    $node->created = time();
    $node->changed = $node->created;
    $node->promote = 0; // Display on front page ? 1 : 0
    $node->sticky = 0;  // Display top of page ? 1 : 0
    $node->format = 2;  // 1:Filtered HTML, 2: Full HTML
    $node->status = 1;   // Published ? 1 : 0
    $node->language = 'en';
    
    $node->query_tools['#new'] = TRUE;
    
    $this->items[] = $node;
  }
}